import shutil
from distutils import log as distutils_log
from pathlib import Path
from typing import Any, Dict

import skbuild
import skbuild.constants

__all__ = ("build",)


def build(setup_kwargs: Dict[str, Any]) -> None:
    """Build Fortran-extensions."""
    skbuild.setup(**setup_kwargs, script_args=["build_ext"])
    src_dir = Path(skbuild.constants.CMAKE_INSTALL_DIR()) / "irilib"
    dest_dir = Path("src/iricore")

    # Delete C-extensions copied in previous runs, just in case.
    remove_files(dest_dir, "**/*.so")

    # Copy built C-extensions back to the project.
    copy_files(src_dir, dest_dir, "**/*.so")


def remove_files(target_dir: Path, pattern: str) -> None:
    """Delete files matched with a glob pattern in a directory tree."""
    for path in target_dir.glob(pattern):
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()
        distutils_log.info(f"removed {path}")


def copy_files(src_dir: Path, dest_dir: Path, pattern: str) -> None:
    """Copy files matched with a glob pattern in a directory tree to another."""
    for src in src_dir.glob(pattern):
        dest = dest_dir / src.relative_to(src_dir)
        if src.is_dir():
            # NOTE: inefficient if subdirectories also match to the pattern.
            copy_files(src, dest, "*")
        else:
            dest.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(src, dest)
            distutils_log.info(f"copied {src} to {dest}")


if __name__ == "__main__":
    build({})



from __future__ import annotations

import warnings
from ctypes import *
from datetime import datetime
from os.path import join as jpath
from typing import Sequence, Literal

import numpy as np
import pymap3d as pm
from numpy.ctypeslib import as_ctypes

from .config import IRI_VERSIONS, DEFAULT_IRI_VERSION
from .iri import iri2016, iri2020, _iri_cfd, iri
from .iri_flags import get_jf
from .modules import srange, R_EARTH


def _clean_ne_for_tec(ne):
    if np.any(np.isnan(ne)) or np.any(np.isinf(ne)):
        warnings.warn("NANs or INFs found in the IRI output; setting them to 0.", stacklevel=2)
    ne = np.where(np.isnan(ne), 0, ne)
    ne = np.where(np.isinf(ne), 0, ne)
    ne_ = ne - np.roll(ne, -1)
    ne = np.where(ne_ > 1e3, np.roll(ne, 1), ne)
    return ne


def _integrate_ne(ne, hstep):
    tec_res = np.sum(ne, axis=-1) * hstep * 1e3 * 1e-16
    if np.any(tec_res < 0) or np.any(tec_res > 100):
        warnings.warn("Unusual TEC value found.", stacklevel=2)
    return tec_res


def vtec(dt: datetime, lat: float | np.ndarray, lon: float | np.ndarray, hbot: float = 90,
         htop: float = 2000, hstep: float = 0.5,
         version: Literal[16, 20] = DEFAULT_IRI_VERSION,
         jf: np.ndarray | str = None):
    """
    Vertical TEC calculated by integrating the electron density on the line of sight. This function
    is not a part of the source IRI code.

    :param dt: time of observation.
    :param lat: Geographical latitude.
    :param lon: Geographical longitude.
    :param hbot: Bottom height limit for integration in [km].
    :param htop: Upper height limit for integration in [km].
    :param hstep: Height step for integration in [km].
    :param version: IRI version number.
    :param jf: Array of JF parameters or string for predefined JF arrays to be used in the IRI_SUB function. See
               :func:`iricore.get_jf` for details. If not specified otherwise, the default IRI JF array will be used.
    :return: Slant TEC.
    """

    if hbot < 60 or htop > 2000:
        raise ValueError("The limits of integration cannot exceed (60, 2000) km.")

    if not isinstance(jf, np.ndarray):
        jf = jf or "default_edens"
        if isinstance(jf, str):
            jf = get_jf(jf)
    else:
        if isinstance(jf, np.ndarray) and jf.size != 50:
            raise ValueError("Length of jf array must be 50")

    npoints = int(np.ceil((htop - hbot) / hstep))
    nstages = int(np.ceil(npoints / 1000))
    stage_ranges = [[
        hbot + hstep * (1000 * i),
        hbot + hstep * (1000 * (i + 1) - 1),
        hstep
    ] for i in range(nstages)]
    stage_ranges[-1][1] = htop

    tec = 0.
    for i in range(nstages):
        iri_res = iri(dt, stage_ranges[i], lat, lon, version, jf)
        tec += _integrate_ne(_clean_ne_for_tec(iri_res.edens), hstep)
    return tec


def stec(el: float, az: float, dt: datetime, lat: float, lon: float, height: float = 0, hbot: float = 90,
         htop: float = 2000, npoints: int = 1000,
         version: Literal[16, 20] = DEFAULT_IRI_VERSION,
         jf: np.ndarray | str = None, _return_ne: bool = False) -> float | Sequence:
    """
    Slant TEC calculated by integrating the electron density on the line of sight. This function
    is not a part of the source IRI code.

    :param el: elevation of observation in [deg].
    :param az: azimuth of observation in [deg].
    :param dt: time of observation.
    :param lat: Geographical latitude.
    :param lon: Geographical longitude.
    :param height: Height of the observer above the sea level.
    :param hbot: Bottom height limit for integration in [km].
    :param htop: Upper height limit for integration in [km].
    :param npoints: Number of points to integrate.
    :param version: IRI version number.
    :param jf: Array of JF parameters or string for predefined JF arrays to be used in the IRI_SUB function. See
               :func:`iricore.get_jf` for details. If not specified otherwise, the default IRI JF array will be used.
    :param _return_ne: If True - also returns IRI Ne output and Ne after and before
                       post-processing (used for debugging).
    :return: Slant TEC.
    """
    if hbot < 60 or htop > 2000:
        raise ValueError("The limits of integration cannot exceed (60, 2000) km.")

    # Calculating input parameters (assuming Earth=sphere)
    hstep = (htop - hbot) / npoints
    heights = np.linspace(hbot, htop, npoints)
    rslant = srange(np.deg2rad(90 - el), heights * 1e3)
    ell = pm.Ellipsoid(R_EARTH, R_EARTH)
    slat, slon, _ = pm.aer2geodetic(az, el, rslant, lat, lon, height, ell=ell)

    if not isinstance(jf, np.ndarray):
        jf = jf or "default_edens"
        if isinstance(jf, str):
            jf = get_jf(jf)
    else:
        if isinstance(jf, np.ndarray) and jf.size != 50:
            raise ValueError("Length of jf array must be 50")

    iri_res = _call_stec(dt, heights, slat, slon, jf, version)
    ne = iri_res[0].transpose()
    ne = ne.reshape((len(heights), -1))[:, 0]

    # Data postprocessing (fixing non-physical IRI output)
    ne_debug = ne.copy()
    ne = _clean_ne_for_tec(ne)
    tec_res = _integrate_ne(ne, hstep)
    if _return_ne:
        return tec_res, ne, ne_debug
    return tec_res


def _call_stec(dt: datetime, heights: Sequence[float], lat: Sequence[float], lon: Sequence[float],
               jf: np.ndarray, version: Literal[16, 20] = DEFAULT_IRI_VERSION):
    if version == 16:
        iricore = iri2016
    elif version == 20:
        iricore = iri2020
    else:
        raise ValueError(f"Available IRI versions: " + ", ".join(IRI_VERSIONS))

    heights = np.asarray(heights)
    jmag = False
    iyyyy = c_int(dt.year)
    mmdd = c_int(100 * dt.month + dt.day)
    dhour = c_float(dt.hour + dt.minute / 60 + dt.second / 3600 + 25.)
    f_heights = as_ctypes(np.array(heights, dtype=np.float32, order="F"))
    f_lat = as_ctypes(np.array(lat, dtype=np.float32, order="F"))
    f_lon = as_ctypes(np.array(lon, dtype=np.float32, order="F"))
    hsize = c_int(len(heights))
    oarr = np.zeros(100, dtype=np.float32, order="F")
    iri_res = np.zeros((20, 1000, len(heights)), dtype=np.float32, order='F')

    datadir = jpath(_iri_cfd, 'data')
    datadir_bytes = bytes(datadir, 'utf-8')

    # aap, af107, nlines = IRI_DATA

    iricore.stec_(as_ctypes(jf), byref(c_bool(jmag)), f_lat, f_lon, f_heights, byref(hsize), byref(iyyyy), byref(mmdd),
                  byref(dhour), as_ctypes(oarr),
                  iri_res.ctypes.data_as(POINTER(c_float)), datadir_bytes, byref(c_int(len(datadir))))

    iri_res = np.ascontiguousarray(iri_res)
    return iri_res



from __future__ import annotations

import numpy as np

R_EARTH = 6378100.


def srange(theta: float | np.ndarray, alt: float | np.ndarray) -> float | np.ndarray:
    """
    :param theta: Zenith angle in [rad].
    :param alt: Altitude in [m].
    :param re: Radius of the Earth in [m].
    :return: Distance in meters from the telescope to the point (theta, alt)
    """
    if isinstance(theta, np.ndarray) and isinstance(alt, np.ndarray):
        raise ValueError("Only one input parameter can be a numpy array.")
    r = -R_EARTH * np.cos(theta) + np.sqrt(
        (R_EARTH * np.cos(theta)) ** 2 + alt ** 2 + 2 * alt * R_EARTH
    )
    return r




from __future__ import annotations

import os
from ctypes import *
from datetime import datetime
from os.path import join as jpath
from typing import Sequence, Literal, Annotated

import numpy as np
from numpy.ctypeslib import as_ctypes

from .config import IRI_VERSIONS, DEFAULT_IRI_VERSION
from .iri_flags import get_jf
from .irioutput import IRIOutput

_iri_cfd = os.path.dirname(os.path.abspath(__file__))

# TODO: Fix data reading from Python
# TODO: Check if data works for all files
# IRI_DATA = readapf107()

try:
    iri2016 = np.ctypeslib.load_library("libiri2016", _iri_cfd)
    iri2020 = np.ctypeslib.load_library("libiri2020", _iri_cfd)
except OSError:
    raise ImportError(
        """
        Could not import IRI libraries, most probably they were not compiled during installation. You can try 
        compiling them manually:
        1) Locate the package installation directory and open it in terminal
        2) Run: 
            >>> cmake .
            >>> make
        """
    )


def iri(dt: datetime, altrange: Annotated[Sequence[float], 3], lat: float | Sequence[float],
        lon: float | Sequence[float], version: Literal[16, 20] = DEFAULT_IRI_VERSION,
        jf: np.ndarray | str = None) -> IRIOutput:
    """
    The main function of the package representing the IRI_SUB routine from the IRI source code. Provides access
    to all IRI calculations through the wide range of parameters.

    :param dt: Date and time - must be a single value.
    :param altrange: Range of altitudes in the form (alt_start, alt_stop, alt_step), all in [km].
    :param lat: Geographical latitude.
    :param lon: Geographical longitude.
    :param version: Version of the IRI to use:

                    * ``version=16`` - IRI-2016
                    * ``version=20`` - IRI-2020

    :param jf: Array of JF parameters or string for predefined JF arrays to be used in the IRI_SUB function. See
               :func:`iricore.get_jf` for details. If not specified otherwise, the default IRI JF array will be used.
    :return: :class:`IRIOutput` dataclass.
    """
    lat = np.atleast_1d(np.asarray(lat))
    lon = np.atleast_1d(np.asarray(lon))
    if not len(lat) == len(lon):
        raise ValueError("Lengths of latitude and longitude arrays must be equal.")

    if not isinstance(jf, np.ndarray):
        jf = jf or "default"
        if isinstance(jf, str):
            jf = get_jf(jf)
    else:
        if isinstance(jf, np.ndarray) and jf.size != 50:
            raise ValueError("Length of jf array must be 50")

    if version == 16:
        iricore = iri2016
    elif version == 20:
        iricore = iri2020
    else:
        raise ValueError(f"Available IRI versions: " + ", ".join(IRI_VERSIONS))

    if (altrange[1] - altrange[0]) / altrange[2] + 1 > 1000:
        raise ValueError("The specified altitude range and step require more that 1000 points for calculation, "
                         "which exceeds IRI limitations. Please consider breaking calculation into chunks.")

    # Converting parameters to Fortran types
    jmag = False
    iyyyy = c_int(dt.year)
    mmdd = c_int(100 * dt.month + dt.day)
    dhour = c_float(dt.hour + dt.minute / 60 + dt.second / 3600 + 25.)
    glat = as_ctypes(np.array(lat, dtype=np.float32, order="F"))
    glon = as_ctypes(np.array(lon, dtype=np.float32, order="F"))
    gsize = c_int(len(lat))
    heibeg = c_float(altrange[0])
    heiend = c_float(altrange[1])
    heistp = c_float(altrange[2])
    oarr = np.zeros(100, dtype=np.float32, order="F")
    iri_res = np.zeros((20, 1000, len(lat)), dtype=np.float32, order='F')

    datadir = jpath(_iri_cfd, 'data')
    datadir_bytes = bytes(datadir, 'utf-8')
    # aap, af107, nlines = IRI_DATA

    # Calling IRI_SUB
    iricore.iricore_(as_ctypes(jf), byref(c_bool(jmag)), glat, glon, byref(gsize), byref(iyyyy), byref(mmdd),
                     byref(dhour), byref(heibeg), byref(heiend), byref(heistp), as_ctypes(oarr),
                     iri_res.ctypes.data_as(POINTER(c_float)), datadir_bytes, byref(c_int(len(datadir))),
                     # aap.ctypes.data_as(POINTER(c_float)), af107.ctypes.data_as(POINTER(c_float)), byref(c_int(nlines))
                     )

    iri_res = np.ascontiguousarray(iri_res)
    return IRIOutput.from_raw(iri_res, lat, lon, altrange)



from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class IRIOutput:
    """
    Dataclass containing :func:`iricore.iri` output. If lat/lon are 1D arrays, then the
    IRI returned values will a 2D array with shape (lat.size, alt.size). If lat/lon
    are a single number, the returned values will be an 1D array with shape (alt.size).

    :param lat: Array of latitudes corresponding to the IRI output (0 axis).
    :param lon: Array of longitudes corresponding to the IRI output (0 axis).
    :param height: Array of heights corresponding to the IRI output (1 axis).
    :param edens: Electron density in [m-3].
    :param ntemp: Neutral temperature in [K].
    :param itemp: Ion temperature in [K].
    :param etemp: Electron temperature in [K].
    :param o: O+ ion density in [%](default) or [m-3].
    :param h: H+ ion density in [%](default) or [m-3].
    :param he: He+ ion density in [%](default) or [m-3].
    :param o2: O2+ ion density in [%](default) or [m-3].
    :param no: NO+ ion density in [%](default) or [m-3].
    :param cluster: Cluster ion density in [%](default) or [m-3].
    :param n: N+ ion density in [%](default) or [m-3].
    :param firi_output: If FIRI was manually selected for the D-region, the ``firi_output`` will be an 1D array
                        containing:

                        * [0:10] standard IRI-Ne for 60,65,..,110km;
                        * [11:21] Friedrich (FIRI) model at these heights;
                        * [22:32] standard Danilov (SW=0, WA=0);
                        * [33:43] for minor Stratospheric Warming (SW=0.5);
                        * [55:65] weak Winter Anomaly (WA=0.5) conditions;
                        * [44:54] for major Stratospheric Warming (SW=1);
                        * [66:76] strong Winter Anomaly (WA=1) conditions.
    """
    lat: np.ndarray
    lon: np.ndarray
    height: np.ndarray
    edens: np.ndarray
    ntemp: np.ndarray
    itemp: np.ndarray
    etemp: np.ndarray
    o: np.ndarray
    h: np.ndarray
    he: np.ndarray
    o2: np.ndarray
    no: np.ndarray
    cluster: np.ndarray
    n: np.ndarray
    firi_output: np.ndarray

    @staticmethod
    def _extract_iri_row(row, arr_shape):
        ncoord, nalt = arr_shape
        row = row.transpose().reshape((ncoord, -1))[:, :nalt]
        return np.squeeze(np.where(row < 0, np.nan, row))

    @classmethod
    def from_raw(cls, output: np.ndarray, lat: np.ndarray, lon: np.ndarray, height_range: [float, float, float]):
        height = np.arange(height_range[0], height_range[1] + height_range[2] / 2, height_range[2])
        arr_shape = (lat.size, height.size)
        return cls(
            lat=lat,
            lon=lon,
            height=height,
            edens=cls._extract_iri_row(output[0], arr_shape),
            ntemp=cls._extract_iri_row(output[1], arr_shape),
            itemp=cls._extract_iri_row(output[2], arr_shape),
            etemp=cls._extract_iri_row(output[3], arr_shape),
            o=cls._extract_iri_row(output[4], arr_shape),
            h=cls._extract_iri_row(output[5], arr_shape),
            he=cls._extract_iri_row(output[6], arr_shape),
            o2=cls._extract_iri_row(output[7], arr_shape),
            no=cls._extract_iri_row(output[8], arr_shape),
            cluster=cls._extract_iri_row(output[9], arr_shape),
            n=cls._extract_iri_row(output[10], arr_shape),
            firi_output=cls._extract_iri_row(output[13], arr_shape),
        )



import os

import fortranformat as ff
import numpy as np

_iri_cfd = os.path.dirname(os.path.abspath(__file__))


def readapf107():
    datadir = os.path.join(_iri_cfd, f"data/")
    with open(os.path.join(datadir, "index/apf107.dat"), "r") as file:
        lines = file.readlines()
    nlines = len(lines)
    linereader = ff.FortranRecordReader("(13I3, 3F5.1)")
    data = np.array([linereader.read(line_) for line_ in lines], order='F', dtype=np.float32)

    f107d = data[:, 13]
    f107_81 = data[:, 14]
    f107_365 = data[:, 15]

    af107 = np.zeros((27000, 3), order='F', dtype=np.float32)
    af107[:nlines, 0] = f107d
    af107[:nlines, 1] = np.where(f107_81 < -4, f107d, f107_81)
    af107[:nlines, 1] = np.where(f107_365 < -4, f107d, f107_365)

    aap = np.zeros((27000, 9), order='F', dtype=np.float32)
    aap[:nlines, :] = data[:, 3:12].astype(np.int32)
    return aap, af107, nlines
